"""
The molgri.plotting sub-package builds on matplotlib and seaborn to provide different visualisations of data based on
a central object that should be represented. For example, SphereGridPlot is initiated with a SphereGridNDim object
and provides various methods to plot and animate the data contained in this object.

All plotting implementations extend basic set-up and saving options set up in the abstract module.
"""